
A ICON MOVEABLE PURE LOADONCALL DISCARDABLE "bwx-console.ico"

/////////////////////////////////////////////////////////////////////////////
//
// RT_MANIFEST
//
CREATEPROCESS_MANIFEST_RESOURCE_ID     RT_MANIFEST    "bwx-console.manifest"
