drop table unsavedfiles;
drop table basefiles;
drop table jobmedia;
drop table file;
drop table job;
drop table media;
drop table client;
drop table pool;
drop table fileset;
drop table path;
drop table filename;
drop table counters;
drop table version;
drop table CDImages;
drop table Device;
drop table Storage;
drop table MediaType;
drop table Status;
drop table MAC;
drop table log;
drop table Location;
drop table locationlog;
