LIBRARY bacula.dll
EXPORTS
 
; compat.o
_Z10open_bpipePciPKc
_Z11close_bpipeP5BPIPE
_Z11close_wpipeP5BPIPE
_Z11strncasecmpPKcS0_i
_Z11win32_cgetsPci
_Z11win32_chdirPKc
_Z11win32_chmodPKct
_Z11win32_fputsPKcP6_iobuf
_Z11win32_mkdirPKc
_Z12UTF8_2_wcharPPcPKc
_Z12gettimeofdayP7timevalP8timezone
_Z12init_signalsPFviE
_Z12wchar_2_UTF8PcPKwi
_Z12wchar_2_UTF8PPcPKw
_Z12win32_getcwdPci
_Z12win32_unlinkPKc
_Z15init_stack_dumpv
_Z15win32_ftruncateix
_Z16wchar_win32_pathPKcPw
_Z17SetVSSPathConvertPFiPKcPciEPFiPKwPwiE
_Z18CreateChildProcessPKcPvS1_S1_
_Z18GetApplicationNamePKcPPcPS0_
_Z18cvt_ftime_to_utimeRK9_FILETIME
_Z18cvt_utime_to_ftimeRKlR9_FILETIME
_Z18unix_name_to_win32PPcS_
_Z21Win32ConvCleanupCachev
_Z28make_win32_path_UTF8_2_wcharPPcPKcPi
_Z4forkv
_Z4killii
_Z4pipePi
_Z4statPKcP4stat
_Z5chownPKcjj
_Z5fcntlii
_Z5fcntliil
_Z5fstatiP4stat
_Z5lstatPKcP4stat
_Z5sleepi
_Z6execvpPKcPPc
_Z6lchownPKcjj
_Z6randomv
_Z7geteuidv
_Z7opendirPKc
_Z7openlogPKcii
_Z7srandomj
_Z7waitpidiPii
_Z8WSA_Initv
_Z8closedirPv
_Z8closelogv
_Z8getArgv0PKc
_Z8getgrgidj
_Z8getpwuidj
_Z8pathconfPKci
_Z8readlinkPKcPci
_Z9ErrorExitPKc
_Z9inet_atonPKcP7in_addr
_Z9nanosleepPK8timespecPS_
_Z9readdir_rPvP6direntPS1_
_ZN6winverC1Ev
_ZN6winverC2Ev
syslog
umask
utime
 
; plugins.o
_Z10new_pluginv
_Z12load_pluginsPvS_PKcS1_PFbP6PluginE
_Z14unload_pluginsv
_Z19dbg_plugin_add_hookPFvP6PluginP6_iobufE

; print.o
_Z10__snprintfPcjPKcz
_Z10__vsprintfPcPKcS_
_Z11__vsnprintfPcjPKcS_
_Z4doprPcjPKcS_PFvS_PjjiE
_Z9__sprintfPcPKcz
 
; winapi.o
_Z17InitWinAPIWrapperv
 
; attribs.o
_Z11decode_statPcP4statiPi
_Z11encode_statPcP4statiii
_Z13decode_LinkFIPcP4stati
_Z14set_attributesP3JCRP4ATTRP5BFILE
_Z16encode_attribsExP3JCRPcP6FF_PKT
_Z18select_data_streamP6FF_PKT
_Z4plugIjEvRT_y
_Z4plugIlEvRT_y
_Z4plugImEvRT_y
_Z4plugIsEvRT_y
_Z4plugItEvRT_y
_Z4plugIyEvRT_y
_Z9win_errorP3JCRPKcm
_Z9win_errorP3JCRPKcPc
 
; bfile.o
_Z11bget_handleP5BFILE
_Z11int32_LE2BEPii
_Z11int64_LE2BEPxx
_Z14have_win32_apiv
_Z15is_win32_streami
_Z15stream_to_asciii
_Z16set_win32_backupP5BFILE
_Z18is_portable_backupP5BFILE
_Z19set_portable_backupP5BFILE
_Z14set_cmd_pluginP5BFILEP3JCR
_Z26processWin32BackupAPIBlockP5BFILEPvl
_Z27is_restore_stream_supportedi
_Z5binitP5BFILE
_Z5bopenP5BFILEPKcit
_Z5breadP5BFILEPvj
_Z6bcloseP5BFILE
_Z6blseekP5BFILExi
_Z6bwriteP5BFILEPvj
_Z8is_bopenP5BFILE
 
; create_file.o
_Z11create_fileP3JCRP4ATTRP5BFILEi
 
; drivetype.o
_Z9drivetypePKcPci
 
; enable_priv.o
_Z24enable_backup_privilegesP3JCRi
 
; find.o
_Z10find_filesP3JCRP6FF_PKTPFiS0_S2_bES4_
_Z13is_in_filesetP6FF_PKT
_Z15init_find_filesv
_Z15term_find_filesP6FF_PKT
_Z16set_find_optionsP6FF_PKTil
_Z22get_win32_drivelettersP6FF_PKTPc
_Z25set_find_changed_functionP6FF_PKTPFbP3JCRS0_E
 
; find_one.o
_Z13check_changesP3JCRP6FF_PKT
_Z13find_one_fileP3JCRP6FF_PKTPFiS0_S2_bEPcjb
_Z13term_find_oneP6FF_PKT
_Z16has_file_changedP3JCRP6FF_PKT
_Z22ff_pkt_set_link_digestP6FF_PKTiPKcj
 
; fstype.o
_Z6fstypePKcPci
 
; match.o
_Z11match_filesP3JCRP6FF_PKTPFiS0_S2_bE
_Z16file_is_excludedP6FF_PKTPKc
_Z16file_is_includedP6FF_PKTPKc
_Z22get_next_included_fileP6FF_PKTP15s_included_file
_Z25add_fname_to_exclude_listP6FF_PKTPKc
_Z25add_fname_to_include_listP6FF_PKTiPKc
_Z26term_include_exclude_filesP6FF_PKT
 
; address_conf.o
_Z14free_addressesP5dlist
_Z15store_addressesP13s_lex_contextP8RES_ITEMii
_Z17get_first_addressP5dlistPci
_Z17sockaddr_get_portPK8sockaddr
_Z17sockaddr_to_asciiPK8sockaddrPci
_Z19build_addresses_strP5dlistPci
_Z20store_addresses_portP13s_lex_contextP8RES_ITEMii
_Z22init_default_addressesPP5dlisti
_Z23store_addresses_addressP13s_lex_contextP8RES_ITEMii
_Z24get_first_port_net_orderP5dlist
_Z25get_first_port_host_orderP5dlist
_Z27sockaddr_get_port_net_orderPK8sockaddr
_ZN6IPADDR11get_addressEPci
_ZN6IPADDR12get_sockaddrEv
_ZN6IPADDR12set_addr_anyEv
_ZN6IPADDR12set_port_netEt
_ZN6IPADDR16get_sockaddr_lenEv
_ZN6IPADDR17build_address_strEPci
_ZN6IPADDR8set_typeENS_6i_typeE
_ZN6IPADDR9copy_addrEPS_
_ZN6IPADDR9set_addr4EP7in_addr
_ZN6IPADDRC1ERKS_
_ZN6IPADDRC1Ei
_ZN6IPADDRC2ERKS_
_ZN6IPADDRC2Ei
_ZNK6IPADDR10get_familyEv
_ZNK6IPADDR18get_port_net_orderEv
_ZNK6IPADDR8get_typeEv
 
; alist.o
_ZN5alist3getEi
_ZN5alist4lastEv
_ZN5alist4nextEv
_ZN5alist4prevEv
_ZN5alist5firstEv
_ZN5alist6appendEPv
_ZN5alist6removeEi
_ZN5alist7destroyEv
_ZN5alist7prependEPv
_ZN5alist9grow_listEv
 
; attr.o
_Z15print_ls_outputP3JCRP4ATTR
_Z24build_attr_output_fnamesP3JCRP4ATTR
_Z24unpack_attributes_recordP3JCRiPciP4ATTR
_Z8new_attrP3JCR
_Z9free_attrP4ATTR
 
; base64.o
_Z11base64_initv
_Z11from_base64PxPc
_Z13bin_to_base64PciS_ii
_Z9to_base64xPc
 
; berrno.o
_ZN6berrno20format_win32_messageEv
_ZN6berrno9bstrerrorEv
 
; bget_msg.o
_Z8bget_msgP5BSOCK
 
; bnet.o
_Z10bnet_closeP5BSOCK
_Z10bnet_fsendP5BSOCKPKcz
_Z10init_bsockP3JCRiPKcS2_iP8sockaddr
_Z10term_bsockP5BSOCK
_Z11read_nbytesP5BSOCKPci
_Z12bnet_connectP3JCRixxPKcPcS3_ii
_Z12is_bnet_stopP5BSOCK
_Z12write_nbytesP5BSOCKPci
_Z13bnet_get_peerP5BSOCKPci
_Z13bnet_strerrorP5BSOCK
_Z13is_bnet_errorP5BSOCK
_Z14bnet_wait_dataP5BSOCKi
_Z15bnet_tls_clientP11TLS_ContextP5BSOCKP5alist
_Z15bnet_tls_serverP11TLS_ContextP5BSOCKP5alist
_Z17bnet_host2ipaddrsPKciPS0_
_Z17bnet_set_blockingP5BSOCK
_Z17bnet_sig_to_asciiP5BSOCK
_Z19bnet_wait_data_intrP5BSOCKi
_Z20bnet_set_buffer_sizeP5BSOCKji
_Z20bnet_set_nonblockingP5BSOCK
_Z21bnet_restore_blockingP5BSOCKi
_Z28bnet_suppress_error_messagesP5BSOCKb
_Z8bnet_sigP5BSOCKi
_Z9bnet_recvP5BSOCK
_Z9bnet_sendP5BSOCK
_Z9dup_bsockP5BSOCK
 
; bnet_server.o
_Z18bnet_thread_serverP5dlistiP9workq_tagPFPvS3_E
_Z23bnet_stop_thread_serverP10pthread_t_
 
; bpipe.o
_Z11run_programPciRS_
_Z23run_program_full_outputPciRS_
 
; breg.o
_Z11new_bregexpPKc
_Z12free_bregexpP7BREGEXP
_Z12get_bregexpsPKc
_Z13free_bregexpsP5alist
_Z14apply_bregexpsPKcP5alistPPc
_Z19bregexp_build_wherePciS_S_S_
_Z21bregexp_escape_stringPcPKcc
_Z28bregexp_get_build_where_sizePcS_S_
_ZN7BREGEXP10edit_substEPKcP12b_regmatch_t
_ZN7BREGEXP12return_fnameEPKci
_ZN7BREGEXP14extract_regexpEPKc
_ZN7BREGEXP16compute_dest_lenEPKcP12b_regmatch_t
_ZN7BREGEXP5debugEv
_ZN7BREGEXP7replaceEPKc
 
; bregex.o
_Z24re_registers_to_regmatchP12re_registersP12b_regmatch_tj
b_re_compile_fastmap
b_re_compile_initialize
b_re_compile_pattern
b_re_match
b_re_search
b_re_set_syntax
b_regcomp
b_regerror
b_regexec
b_regfree
 
; bsnprintf.o
_Z10bvsnprintfPciPKcS_
_Z9bsnprintfPciPKcz
 
; bsock.o
_ZN5BSOCK11set_lockingEv
_Z9new_bsockv
_ZN5BSOCK10free_bsockEv
_ZN5BSOCK12set_blockingEv
_ZN5BSOCK14wait_data_intrEii
_ZN5BSOCK15set_buffer_sizeEji
_ZN5BSOCK15set_nonblockingEv
_ZN5BSOCK16restore_blockingEi
_ZN5BSOCK21authenticate_directorEPKcS1_P11TLS_ContextPci
_ZN5BSOCK4initEv
_ZN5BSOCK4recvEv
_ZN5BSOCK4sendEv
_ZN5BSOCK5_openEP3JCRPKcPcS4_ixPi
_ZN5BSOCK5closeEv
_ZN5BSOCK5fsendEPKcz
_ZN5BSOCK6signalEi
_ZN5BSOCK7connectEP3JCRixxPKcPcS4_ii
_ZN5BSOCK7despoolEPFvlEl
_ZN5BSOCK7destroyEv
_ZN5BSOCK8fin_initEP3JCRiPKcS3_iP8sockaddr
_ZN5BSOCK8get_peerEPci
_ZN5BSOCK9bstrerrorEv
_ZN5BSOCK9wait_dataEii
_ZN5BSOCK18set_source_addressEP5dlist
 
; bsys.o
_Z10b_strerroriPcj
_Z11bmicrosleepii
_Z15create_pid_filePcPKci
_Z15delete_pid_filePcPKci
_Z15escape_filenamePKc
_Z15read_state_filePcPKci
_Z16write_state_filePcPKci
_Z20make_unique_filenamePPciS_
_Z5bfreePv
_Z6bfgetsPciP6_iobuf
_Z7bcallocjj
_Z7bstrcmpPKcS0_
_Z7cstrlenPKc
_Z8b_mallocPKcij
_Z8breallocPvj
_Z8bstrncatPcPKci
_Z8bstrncatPcR8POOL_MEMi
_Z8bstrncpyPcPKci
_Z8bstrncpyPcR8POOL_MEMi

; binflate.o
_Z8ZdeflatePciS_Ri
_Z8ZinflatePciS_Ri
 
; btime.o
_Z10bstrftimesPcix
_Z11date_decodedPjPhS0_
_Z11date_encodejhh
_Z11time_decodedPhS_S_Pf
_Z11time_encodehhhf
_Z12bstrftime_ncPcix
_Z12bstrftime_nyPcix
_Z12str_to_utimePc
_Z13btime_to_unixx
_Z14btime_to_utimex
_Z16date_time_decodeP9date_timePjPhS2_S2_S2_S2_Pf
_Z16date_time_encodeP9date_timejhhhhhf
_Z16get_current_timeP9date_time
_Z17date_time_compareP9date_timeS0_
_Z17get_current_btimev
_Z6tm_womii
_Z6tm_woyl
_Z9bstrftimePcix
_Z9bstrutimePcix
_Z9tm_decodeP9date_timeP2tm
_Z9tm_encodeP9date_timeP2tm
 
; btimers.o
_Z16stop_bsock_timerP8btimer_t
_Z16stop_child_timerP8btimer_t
_Z17start_bsock_timerP5BSOCKj
_Z17start_child_timerP3JCRij
_Z17stop_thread_timerP8btimer_t
_Z18start_thread_timerP3JCRP10pthread_t_j

; cram-md5.o
_Z16cram_md5_respondP5BSOCKPKcPiS3_
_Z18cram_md5_challengeP5BSOCKPKcii
 
; crc32.o
_Z6bcrc32Phi
 
; crypto.o
_Z11init_cryptov
_Z13CryptoData_itv
_Z13SignerInfo_itv
_Z14CryptoData_newv
_Z14SignerInfo_newv
_Z14cleanup_cryptov
_Z14d2i_CryptoDataPP10CryptoDataPPKhl
_Z14d2i_SignerInfoPP10SignerInfoPPKhl
_Z14i2d_CryptoDataP10CryptoDataPPh
_Z14i2d_SignerInfoP10SignerInfoPPh
_Z15CryptoData_freeP10CryptoData
_Z15SignerInfo_freeP10SignerInfo
_Z15crypto_sign_newP3JCR
_Z15crypto_strerror14crypto_error_t
_Z16RecipientInfo_itv
_Z16SignatureData_itv
_Z16crypto_sign_freeP9Signature
_Z17RecipientInfo_newv
_Z17SignatureData_newv
_Z17crypto_cipher_newP14Crypto_SessionbPj
_Z17crypto_digest_newP3JCR15crypto_digest_t
_Z17d2i_RecipientInfoPP13RecipientInfoPPKhl
_Z17d2i_SignatureDataPP13SignatureDataPPKhl
_Z17i2d_RecipientInfoP13RecipientInfoPPh
_Z17i2d_SignatureDataP13SignatureDataPPh
_Z18RecipientInfo_freeP13RecipientInfo
_Z18SignatureData_freeP13SignatureData
_Z18crypto_cipher_freeP14Cipher_Context
_Z18crypto_digest_freeP6Digest
_Z18crypto_digest_nameP6Digest
_Z18crypto_keypair_dupP12X509_Keypair
_Z18crypto_keypair_newv
_Z18crypto_session_new15crypto_cipher_tP5alist
_Z18crypto_sign_decodeP3JCRPKhj
_Z18crypto_sign_encodeP9SignaturePhPj
_Z18crypto_sign_verifyP9SignatureP12X509_KeypairP6Digest
_Z19crypto_keypair_freeP12X509_Keypair
_Z19crypto_session_freeP14Crypto_Session
_Z20crypto_cipher_updateP14Cipher_ContextPKhjS2_Pj
_Z20crypto_digest_updateP6DigestPKhj
_Z21crypto_session_decodePKhjP5alistPP14Crypto_Session
_Z21crypto_session_encodeP14Crypto_SessionPhPj
_Z22crypto_cipher_finalizeP14Cipher_ContextPhPj
_Z22crypto_digest_finalizeP6DigestPhPj
_Z22crypto_keypair_has_keyPKc
_Z22crypto_sign_add_signerP9SignatureP6DigestP12X509_Keypair
_Z22crypto_sign_get_digestP9SignatureP12X509_KeypairR15crypto_digest_tPP6Digest
_Z23crypto_keypair_load_keyP12X509_KeypairPKcPFiPciPKvES5_
_Z24crypto_keypair_load_certP12X509_KeypairPKc
_Z25crypto_digest_stream_typei
_Z27crypto_default_pem_callbackPciPKv
 
; daemon.o
_Z12daemon_startv
 
; dlist.o
_Z15new_dlistStringPKc
_Z15new_dlistStringPKci
_ZN5dlist12insert_afterEPvS0_
_ZN5dlist13binary_insertEPvPFiS0_S0_E
_ZN5dlist13binary_searchEPvPFiS0_S0_E
_ZN5dlist13insert_beforeEPvS0_
_ZN5dlist22binary_insert_multipleEPvPFiS0_S0_E
_ZN5dlist4nextEPv
_ZN5dlist4prevEPv
_ZN5dlist6appendEPv
_ZN5dlist6removeEPv
_ZN5dlist7destroyEv
_ZN5dlist7prependEPv
 
; edit.o
_Z16is_a_number_listPKc
_Z10add_commasPcS_
_Z10edit_int64xPc
_Z10edit_utimexPci
_Z11edit_uint64yPc
_Z11is_a_numberPKc
_Z12str_to_int64Pc
_Z13is_an_integerPKc
_Z13is_name_validPKcPPc
_Z13str_to_uint64Pc
_Z14size_to_uint64PciPy
_Z15speed_to_uint64PciPy
_Z17duration_to_utimePcPx
_Z22edit_int64_with_commasxPc
_Z23edit_uint64_with_commasyPc
_Z23edit_uint64_with_suffixyPc
 
; fnmatch.o
fnmatch

; guid_to_name.o
_ZN9guid_list11uid_to_nameEjPci
_ZN9guid_list11gid_to_nameEjPci
_Z14free_guid_listP9guid_list
_Z13new_guid_listv

; hmac.o
_Z8hmac_md5PhiS_iS_
 
; htable.o
_ZN6htable11hash_mallocEi
_ZN6htable10grow_tableEv
_ZN6htable10hash_indexEPc
_ZN6htable4initEPvS0_ii
_ZN6htable4nextEv
_ZN6htable4sizeEv
_ZN6htable5firstEv
_ZN6htable5statsEv
_ZN6htable6insertEPcPv
_ZN6htable6lookupEPc
_ZN6htable7destroyEv
_ZN6htableC1EPvS0_ii
_ZN6htableC2EPvS0_ii
    
 
; jcr.o
_Z10b_free_jcrPKciP3JCR
_Z11unlock_jobsv
_Z12jcr_walk_endP3JCR
_Z12job_end_pushP3JCRPFvS0_PvES1_
_Z13get_jcr_by_idj
_Z13jcr_walk_nextP3JCR
_Z14jcr_walk_startv
_Z9job_countv
_Z14set_jcr_in_tsdP3JCR
_Z16get_jcr_from_tsdv
_Z18get_jobid_from_tsdv
_Z18get_jcr_by_sessionjj
_Z18init_jcr_subsystemv
_ZN3JCR12setJobStatusEi
_Z19init_last_jobs_listv
_Z19lock_last_jobs_listv
_Z19read_last_jobs_listiy
_Z19term_last_jobs_listv
_Z20get_jcr_by_full_namePc
_Z20write_last_jobs_listiy
_Z21unlock_last_jobs_listv
_Z23get_jcr_by_partial_namePc
_Z24get_next_jobid_from_listPPcPj
_Z7new_jcriPFvP3JCRE
_Z9lock_jobsv
_ZN3JCR14get_ActionNameEb
_ZN3JCR17get_OperationNameEv
_ZN3JCR21my_thread_send_signalEi
_ZN3JCR8JobReadsEv
_Z16dbg_jcr_add_hookPFvP3JCRP6_iobufE
timeout_handler


 
; lex.o
_Z11scan_to_eolP13s_lex_context
_Z12lex_get_charP13s_lex_context
_Z13lex_get_tokenP13s_lex_contexti
_Z13lex_open_fileP13s_lex_contextPKcPFvS2_iS0_S2_zE
_Z14lex_close_fileP13s_lex_context
_Z14lex_tok_to_stri
_Z14lex_unget_charP13s_lex_context
_Z20scan_to_next_not_eolP13s_lex_context
_Z29lex_set_default_error_handlerP13s_lex_context
_Z32lex_set_error_handler_error_typeP13s_lex_contexti
 
; md5.o
_Z12MD5TransformPjS_
_Z7MD5InitP10MD5Context
_Z8MD5FinalPhP10MD5Context
_Z9MD5UpdateP10MD5ContextPhj
 
; mem_pool.o
_Z13sm_get_memoryPKcii
_Z17close_memory_poolv
_Z18sm_get_pool_memoryPKcii
_Z19sm_free_pool_memoryPKciPc
_Z21sm_sizeof_pool_memoryPKciPc
_Z22sm_realloc_pool_memoryPKciPci
_Z23print_memory_pool_statsv
_Z25sm_check_pool_memory_sizePKciPci
_Z27garbage_collect_memory_poolv
_Z9pm_strcatPPcPKc
_Z9pm_strcatR8POOL_MEMPKc
_Z9pm_strcatRPcPKc
_Z9pm_strcatRPcR8POOL_MEM
_Z9pm_strcpyPPcPKc
_Z9pm_strcpyR8POOL_MEMPKc
_Z9pm_strcpyRPcPKc
_Z9pm_strcpyRPcR8POOL_MEM
_Z9pm_memcpyPPcPKci
_Z9pm_memcpyR8POOL_MEMPKci
_Z9pm_memcpyRPcPKci
_Z9pm_memcpyRPcR8POOL_MEMi
_ZN8POOL_MEM10realloc_pmEi
_ZN8POOL_MEM6strcatEPKc
_ZN8POOL_MEM6strcpyEPKc
_ZN8POOL_MEM8max_sizeEv
 
; message.o
_Z10get_hangupv
_Z10set_hangupi
_Z10my_name_isiPPcPKc
_Z12add_msg_destP4MSGSiiPcS1_
_Z12rem_msg_destP4MSGSiiPc
_Z13free_msgs_resP4MSGS
_Z16dequeue_messagesP3JCR
_Z16dispatch_messageP3JCRixPc
_Z16init_console_msgPKc
_Z25register_message_callbackPFviPcE
_Z4JmsgP3JCRixPKcz
_Z4MmsgPPcPKcz
_Z4MmsgR8POOL_MEMPKcz
_Z4MmsgRPcPKcz
_Z4QmsgP3JCRixPKcz
_Z5d_msgPKciiS0_z
_Z5e_msgPKciiiS0_z
_Z5j_msgPKciP3JCRixS0_z
_Z5m_msgPKciPPcS0_z
_Z5m_msgPKciRPcS0_z
_Z5p_msgPKciiS0_z
_Z5q_msgPKciP3JCRixS0_z
_Z5t_msgPKciiS0_z
_Z8init_msgP3JCRP4MSGS
_Z8term_msgv
_Z9close_msgP3JCR
_Z9get_tracev
_Z9set_tracei

; mkpath.o
_Z13path_list_addP3JCRjPc
_Z14free_path_listP3JCR
_Z16path_list_lookupP3JCRPc
_Z8makepathP4ATTRPKcttjji

; mntent_cache.o
_Z18flush_mntent_cachev
_Z19find_mntent_mappingj

; priv.o
_Z4dropPcS_b

; pythonlib.o
_Z21generate_daemon_eventP3JCRPKc
 
; queue.o
_Z5qnextP7b_queueS0_
_Z7qdchainP7b_queue
_Z7qinsertP7b_queueS0_
_Z7qremoveP7b_queue
 
; rblist.o
_ZN6rblist11left_rotateEPv
_ZN6rblist12right_rotateEPv
_ZN6rblist3anyEPv
_ZN6rblist4nextEPv
_ZN6rblist5firstEv
_ZN6rblist6insertEPvPFiS0_S0_E
_ZN6rblist6removeEPv
_ZN6rblist6searchEPvPFiS0_S0_E
_ZN6rblist7destroyEv
 
; runscript.o
_Z11run_scriptsP3JCRP5alistPKc
_Z13new_runscriptv
_Z14copy_runscriptP9RUNSCRIPT
_Z14free_runscriptP9RUNSCRIPT
_Z15free_runscriptsP5alist
_ZN9RUNSCRIPT10set_targetEPKc
_ZN9RUNSCRIPT11set_commandEPKci
_ZN9RUNSCRIPT13reset_defaultEb
_ZN9RUNSCRIPT21set_job_code_callbackEPFPcP3JCRPKcE
_ZN9RUNSCRIPT3runEP3JCRPKc
_ZN9RUNSCRIPT5debugEv
_ZN9RUNSCRIPT8is_localEv

 
; rwlock.o
_Z11rwl_destroyP12s_rwlock_tag
_Z12rwl_readlockP12s_rwlock_tag
_Z15rwl_writelock_pP12s_rwlock_tagPKci
_Z14rwl_readunlockP12s_rwlock_tag
_Z15rwl_readtrylockP12s_rwlock_tag
_Z15rwl_writeunlockP12s_rwlock_tag
_Z16rwl_writetrylockP12s_rwlock_tag
_Z8rwl_initP12s_rwlock_tagi
 
; scan.o
_Z10parse_argsPcPS_PiS0_S0_i
_Z11skip_spacesPPc
_Z14skip_nonspacesPPc
_Z15parse_args_onlyPcPS_PiS0_S0_i
_Z19strip_leading_spacePc
_Z19strip_trailing_junkPc
_Z22strip_trailing_newlinePc
_Z22strip_trailing_slashesPc
_Z23split_path_and_filenamePKcPPcPiS2_S3_
_Z7bsscanfPKcS0_z
_Z7fstrschPKcS0_
_Z8next_argPPc
 
; serial.o
_Z12serial_btimePPhx
_Z12serial_int16PPhs
_Z12serial_int32PPhi
_Z12serial_int64PPhx
_Z13serial_stringPPhPKc
_Z13serial_uint16PPht
_Z13serial_uint32PPhj
_Z13serial_uint64PPhy
_Z14serial_float64PPhd
_Z14unserial_btimePPh
_Z14unserial_int16PPh
_Z14unserial_int32PPh
_Z15unserial_uint16PPh
_Z15unserial_uint32PPh
_Z15unserial_uint64PPh
_Z16unserial_float64PPh
_Z15unserial_stringPPhPci
 
; smartall.o
_Z10sm_reallocPKciPvj
_Z12actuallyfreePv
_Z12sm_check_rtnPKcib
_Z12sm_new_ownerPKciPc
_Z14actuallycallocjj
_Z14actuallymallocj
_Z15actuallyreallocPvj
_Z7sm_dumpbb
_Z7sm_freePKciPv
_Z8sm_checkPKcib
_Z9sm_callocPKcijj
_Z9sm_mallocPKcij
_Z9sm_staticb
 
; tls.o
_ZN5BSOCK8free_tlsEv
_Z14get_tls_enableP11TLS_Context
_Z15get_tls_requireP11TLS_Context
_Z15new_tls_contextPKcS0_S0_S0_PFiPciPKvES3_S0_b
_Z15tls_bsock_readnP5BSOCKPci
_Z16free_tls_contextP11TLS_Context
_Z16tls_bsock_acceptP5BSOCK
_Z16tls_bsock_writenP5BSOCKPci
_Z17tls_bsock_connectP5BSOCK
_Z18new_tls_connectionP11TLS_Contexti
_Z18tls_bsock_shutdownP5BSOCK
_Z19free_tls_connectionP14TLS_Connection
_Z25tls_postconnect_verify_cnP3JCRP14TLS_ConnectionP5alist
_Z27tls_postconnect_verify_hostP3JCRP14TLS_ConnectionPKc
 
; tree.o
_Z11tree_relcwdPcP11s_tree_rootP11s_tree_node
_Z12tree_getpathP11s_tree_nodePci
_Z14make_tree_pathPcP11s_tree_root
_Z16insert_tree_nodePcS_iP11s_tree_rootP11s_tree_node
_Z8new_treei
_Z8tree_cwdPcP11s_tree_rootP11s_tree_node
_Z9free_treeP11s_tree_root
 
; util.o
_Z11bash_spacesPc
_Z11bash_spacesR8POOL_MEM
_Z11encode_modetPc
_Z11encode_timexPc
_Z11is_buf_zeroPci
_Z13unbash_spacesPc
_Z13unbash_spacesR8POOL_MEM
_Z14edit_job_codesP3JCRPcS1_PKcPFS1_S0_S3_E
_Z15job_type_to_stri
_Z16job_level_to_stri
_Z16make_session_keyPcS_i
_Z17job_status_to_stri
_Z18do_shell_expansionPci
_Z18jobstatus_to_asciiiPci
_Z19last_path_separatorPKc
_Z21set_working_directoryPc
_Z5lcasePc
_Z20volume_status_to_strPKc 
_Z22jobstatus_to_ascii_guiiPci

; var.o
_Z10var_configP6var_st12var_config_tz
_Z10var_createPP6var_st
_Z10var_expandP6var_stPKciPPcPii
_Z10var_formatP6var_stPPciPKcz
_Z11var_destroyP6var_st
_Z11var_formatvP6var_stPPciPKcS1_
_Z12var_strerrorP6var_st8var_rc_t
_Z12var_unescapeP6var_stPKciPcii
 
; watchdog.o
_Z12new_watchdogv
_Z13stop_watchdogv
_Z14start_watchdogv
_Z17register_watchdogP12s_watchdog_t
_Z19unregister_watchdogP12s_watchdog_t
watchdog_thread

; lockmgr.o
_Z6lmgr_pPP16pthread_mutex_t_
_Z6lmgr_vPP16pthread_mutex_t_

;_Z15lmgr_mutex_lockPP16pthread_mutex_t_PKci
;_Z17lmgr_mutex_unlockPP16pthread_mutex_t_PKci
;_Z14lmgr_cond_waitPP15pthread_cond_t_PP16pthread_mutex_t_
;_Z18lmgr_thread_createPP10pthread_t_PKP15pthread_attr_t_PFPvS6_ES6_
;_Z17lmgr_cleanup_mainv

console_command DATA
bplugin_list DATA
plugin_bopen DATA
plugin_bclose DATA
plugin_bwrite DATA
plugin_bread DATA
plugin_blseek DATA
exepath DATA
beef DATA
version DATA
dist_name DATA
