LIBRARY bacula_cats.dll
EXPORTS

; bvfs.o
_Z15bvfs_parent_dirPc
_Z17bvfs_basename_dirPc
_Z17bvfs_update_cacheP3JCRP4B_DB
_Z32bvfs_update_path_hierarchy_cacheP3JCRP4B_DBPc
_ZN4Bvfs12_handle_pathEPviPPc
_ZN4Bvfs12update_cacheEv
_ZN4Bvfs15ls_special_dirsEv
_ZN4Bvfs18get_dir_filenameidEv
_ZN4Bvfs21get_all_file_versionsEjjPKc
_ZN4Bvfs6ch_dirEPKc
_ZN4Bvfs7ls_dirsEv
_ZN4Bvfs8get_rootEv
_ZN4Bvfs8ls_filesEv
_ZN4BvfsC1EP3JCRP4B_DB
_ZN4BvfsC2EP3JCRP4B_DB
_ZN4BvfsD0Ev
_ZN4BvfsD1Ev
_ZN4BvfsD2Ev

 
; sql.o
_Z10_db_unlockPKciP4B_DB
_Z11db_strerrorP4B_DB
_Z11list_dashesP4B_DBPFvPvPKcES1_
_Z11list_resultP3JCRP4B_DBPFvPvPKcES3_11e_list_type
_Z14db_debug_printP3JCRP6_iobuf
_Z16db_int64_handlerPviPPc
_Z15db_list_handlerPviPPc
_Z18db_end_transactionP3JCRP4B_DB
_Z18get_sql_record_maxP3JCRP4B_DB
_Z19split_path_and_fileP3JCRP4B_DBPKc
_Z20check_tables_versionP3JCRP4B_DB
_Z20db_start_transactionP3JCRP4B_DB
_Z23db_open_batch_connexionP3JCRP4B_DB
_Z24db_check_max_connectionsP3JCRP4B_DBj
_Z28db_check_backend_thread_safev
_Z7QueryDBPKciP3JCRP4B_DBPc
_Z8DeleteDBPKciP3JCRP4B_DBPc
_Z8InsertDBPKciP3JCRP4B_DBPc
_Z8UpdateDBPKciP3JCRP4B_DBPc
_Z8_db_lockPKciP4B_DB
_ZN9dbid_listC1Ev
_ZN9dbid_listC2Ev
_ZN9dbid_listD1Ev
_ZN9dbid_listD2Ev
 
; sql_create.o
_Z20db_create_job_recordP3JCRP4B_DBP7JOB_DBR
_Z21db_create_pool_recordP3JCRP4B_DBP8POOL_DBR
_Z22db_create_media_recordP3JCRP4B_DBP9MEDIA_DBR
_Z23db_create_client_recordP3JCRP4B_DBP10CLIENT_DBR
_Z23db_create_device_recordP3JCRP4B_DBP10DEVICE_DBR
_Z24db_create_base_file_listP3JCRP4B_DBPc
_Z24db_create_counter_recordP3JCRP4B_DBP11COUNTER_DBR
_Z24db_create_fileset_recordP3JCRP4B_DBP11FILESET_DBR
_Z24db_create_storage_recordP3JCRP4B_DBP11STORAGE_DBR
_Z25db_create_jobmedia_recordP3JCRP4B_DBP12JOBMEDIA_DBR
_Z26db_create_mediatype_recordP3JCRP4B_DBP13MEDIATYPE_DBR
_Z27db_create_attributes_recordP3JCRP4B_DBP8ATTR_DBR
_Z27db_write_batch_file_recordsP3JCR
_Z32db_create_file_attributes_recordP3JCRP4B_DBP8ATTR_DBR
_Z37db_commit_base_file_attributes_recordP3JCRP4B_DB
_Z37db_create_base_file_attributes_recordP3JCRP4B_DBP8ATTR_DBR
 
; sql_delete.o
_Z21db_delete_pool_recordP3JCRP4B_DBP8POOL_DBR
_Z21db_purge_media_recordP3JCRP4B_DBP9MEDIA_DBR
_Z22db_delete_media_recordP3JCRP4B_DBP9MEDIA_DBR
 
; sql_find.o
_Z18db_find_last_jobidP3JCRP4B_DBPKcP7JOB_DBR
_Z19db_find_next_volumeP3JCRP4B_DBibP9MEDIA_DBR
_Z22db_find_job_start_timeP3JCRP4B_DBP7JOB_DBRPPc
_Z24db_find_failed_job_sinceP3JCRP4B_DBP7JOB_DBRPcRi
_Z27db_find_last_job_start_timeP3JCRP4B_DBP7JOB_DBRPPci
 
; sql_get.o
_Z15db_get_pool_idsP3JCRP4B_DBPiPPj
_Z16db_get_file_listP3JCRP4B_DBPcPFiPviPS3_ES4_
_Z16db_get_media_idsP3JCRP4B_DBP9MEDIA_DBRPiPPj
_Z17db_get_base_jobidP3JCRP4B_DBP7JOB_DBRPj
_Z17db_get_client_idsP3JCRP4B_DBPiPPj
_Z17db_get_job_recordP3JCRP4B_DBP7JOB_DBR
_Z18db_get_pool_recordP3JCRP4B_DBP8POOL_DBR
_Z18db_get_query_dbidsP3JCRP4B_DBR8POOL_MEMR9dbid_list
_Z19db_get_media_recordP3JCRP4B_DBP9MEDIA_DBR
_Z20db_get_client_recordP3JCRP4B_DBP10CLIENT_DBR
_Z21db_get_base_file_listP3JCRP4B_DBPFiPviPPcES3_
_Z21db_get_counter_recordP3JCRP4B_DBP11COUNTER_DBR
_Z21db_get_fileset_recordP3JCRP4B_DBP11FILESET_DBR
_Z22db_accurate_get_jobidsP3JCRP4B_DBP7JOB_DBRP11db_list_ctx
_Z23db_get_job_volume_namesP3JCRP4B_DBjPPc
_Z23db_get_num_pool_recordsP3JCRP4B_DB
_Z23db_get_used_base_jobidsP3JCRP4B_DBPcP11db_list_ctx
_Z24db_get_num_media_recordsP3JCRP4B_DB
_Z28db_get_job_volume_parametersP3JCRP4B_DBjPP10VOL_PARAMS
_Z29db_get_file_attributes_recordP3JCRP4B_DBPcP7JOB_DBRP8FILE_DBR
 
; sql_list.o
_Z22db_list_copies_recordsP3JCRP4B_DBjPcPFvPvPKcES4_11e_list_type
_Z17db_list_sql_queryP3JCRP4B_DBPKcPFvPvS4_ES5_i11e_list_type
_Z18db_list_job_totalsP3JCRP4B_DBP7JOB_DBRPFvPvPKcES5_
_Z19db_list_job_recordsP3JCRP4B_DBP7JOB_DBRPFvPvPKcES5_11e_list_type
_Z20db_list_pool_recordsP3JCRP4B_DBP8POOL_DBRPFvPvPKcES5_11e_list_type
_Z21db_list_files_for_jobP3JCRP4B_DBjPFvPvPKcES3_
_Z21db_list_media_recordsP3JCRP4B_DBP9MEDIA_DBRPFvPvPKcES5_11e_list_type
_Z22db_list_client_recordsP3JCRP4B_DBPFvPvPKcES3_11e_list_type
_Z22db_list_joblog_recordsP3JCRP4B_DBjPFvPvPKcES3_11e_list_type
_Z24db_list_jobmedia_recordsP3JCRP4B_DBjPFvPvPKcES3_11e_list_type
_Z26db_list_base_files_for_jobP3JCRP4B_DBjPFvPvPKcES3_
 
; sql_update.o
_Z15db_update_statsP3JCRP4B_DBx
_Z19db_mark_file_recordP3JCRP4B_DByj
_Z21db_update_pool_recordP3JCRP4B_DBP8POOL_DBR
_Z22db_update_media_recordP3JCRP4B_DBP9MEDIA_DBR
_Z23db_update_client_recordP3JCRP4B_DBP10CLIENT_DBR
_Z24db_make_inchanger_uniqueP3JCRP4B_DBP9MEDIA_DBR
_Z24db_update_counter_recordP3JCRP4B_DBP11COUNTER_DBR
_Z24db_update_job_end_recordP3JCRP4B_DBP7JOB_DBR
_Z24db_update_media_defaultsP3JCRP4B_DBP9MEDIA_DBR
_Z24db_update_storage_recordP3JCRP4B_DBP11STORAGE_DBR
_Z26db_update_job_start_recordP3JCRP4B_DBP7JOB_DBR
_Z28db_add_digest_to_file_recordP3JCRP4B_DByPci
 
_Z11db_get_typev
_Z12db_sql_queryP4B_DBPKcPFiPviPPcES3_
_Z16db_escape_stringP3JCRP4B_DBPcS3_i
_Z16db_init_databaseP3JCRPKcS2_S2_S2_iS2_i
_Z16db_open_databaseP3JCRP4B_DB
_Z17db_close_databaseP3JCRP4B_DB
_Z17db_thread_cleanupv

client_backups DATA
list_pool DATA
drop_deltabs DATA
create_deltabs DATA
create_delindex DATA
insert_delcand DATA
select_backup_del DATA
select_verify_del DATA
select_restore_del DATA
select_admin_del DATA
select_migrate_del DATA
select_copy_del DATA
select_job DATA
count_select_job DATA
del_File DATA
cnt_DelCand DATA
del_Job DATA
del_JobMedia DATA
cnt_JobMedia DATA
sel_JobMedia DATA
upd_Purged DATA
uar_list_jobs DATA
uar_print_jobs DATA
uar_file DATA
uar_count_files DATA
uar_sel_files DATA
uar_del_temp DATA
uar_del_temp1 DATA
uar_create_temp DATA
uar_create_temp1 DATA
uar_last_full DATA
uar_full DATA
uar_inc DATA
uar_list_temp DATA
uar_sel_all_temp1 DATA
uar_sel_fileset DATA
uar_mediatype DATA
uar_jobid_fileindex DATA
uar_dif DATA
uar_sel_all_temp DATA
uar_count_files DATA
uar_jobids_fileindex DATA
uar_jobid_fileindex_from_dir DATA
uar_jobid_fileindex_from_table DATA
uar_sel_jobid_temp DATA
cleanup_created_job DATA
cleanup_running_job DATA
